@@ -1,12 +1,14 @@
 // SPDX-License-Identifier: GPL-2.0-only
 /*
- * Copyright (c) 2013 - 2017 DisplayLink (UK) Ltd.
+ * Copyright (c) 2013 - 2019 DisplayLink (UK) Ltd.
  *
  * This file is subject to the terms and conditions of the GNU General Public
  * License v2. See the file COPYING in the main directory of this archive for
  * more details.
  */
 
+#include "linux/thread_info.h"
+#include "linux/mm.h"
 #include <drm/drmP.h>
 #include <drm/drm_edid.h>
 #include "evdi_drm.h"
@@ -16,6 +18,17 @@
 #include <linux/mutex.h>
 #include <linux/compiler.h>
 
+#if KERNEL_VERSION(4, 12, 0) > LINUX_VERSION_CODE
+static inline void drm_framebuffer_put(struct drm_framebuffer *fb)
+{
+	drm_framebuffer_unreference(fb);
+}
+static inline void drm_framebuffer_get(struct drm_framebuffer *fb)
+{
+	drm_framebuffer_reference(fb);
+}
+#endif
+
 struct evdi_event_cursor_set_pending {
 	struct drm_pending_event base;
 	struct drm_evdi_event_cursor_set cursor_set;
@@ -56,15 +69,14 @@ struct evdi_painter {
 	unsigned int edid_length;
 
 	struct mutex lock;
-	struct mutex new_scanout_fb_lock;
 	struct drm_clip_rect dirty_rects[MAX_DIRTS];
 	int num_dirts;
-	struct evdi_framebuffer *new_scanout_fb;
 	struct evdi_framebuffer *scanout_fb;
 
 	struct drm_file *drm_filp;
 
 	bool was_update_requested;
+	bool needs_full_modeset;
 };
 
 static void expand_rect(struct drm_clip_rect *a, const struct drm_clip_rect *b)
@@ -119,14 +131,14 @@ static void collapse_dirty_rects(struct drm_clip_rect *rects, int *count)
 	*count = 1;
 }
 
-static int copy_primary_pixels(struct evdi_framebuffer *ufb,
-			char __user *buffer,
-			int buf_byte_stride,
-			int num_rects, struct drm_clip_rect *rects,
-			int const max_x,
-			int const max_y)
+static int copy_primary_pixels(struct evdi_framebuffer *efb,
+			       char __user *buffer,
+			       int buf_byte_stride,
+			       int num_rects, struct drm_clip_rect *rects,
+			       int const max_x,
+			       int const max_y)
 {
-	struct drm_framebuffer *fb = &ufb->base;
+	struct drm_framebuffer *fb = &efb->base;
 	struct drm_clip_rect *r;
 
 	EVDI_CHECKPT();
@@ -135,7 +147,7 @@ static int copy_primary_pixels(struct evdi_framebuffer *ufb,
 		const int byte_offset = r->x1 * 4;
 		const int byte_span = (r->x2 - r->x1) * 4;
 		const int src_offset = fb->pitches[0] * r->y1 + byte_offset;
-		const char *src = (char *)ufb->obj->vmapping + src_offset;
+		const char *src = (char *)efb->obj->vmapping + src_offset;
 		const int dst_offset = buf_byte_stride * r->y1 + byte_offset;
 		char __user *dst = buffer + dst_offset;
 		int y = r->y2 - r->y1;
@@ -161,22 +173,21 @@ static int copy_primary_pixels(struct evdi_framebuffer *ufb,
 	return 0;
 }
 
-static int copy_cursor_pixels(struct evdi_framebuffer *ufb,
-			char __user *buffer,
-			int buf_byte_stride,
-			struct evdi_cursor *cursor)
+static void copy_cursor_pixels(struct evdi_framebuffer *efb,
+			       char __user *buffer,
+			       int buf_byte_stride,
+			       struct evdi_cursor *cursor)
 {
-	int ret = 0;
-
 	if (evdi_enable_cursor_blending) {
 		evdi_cursor_lock(cursor);
-		ret = evdi_cursor_compose_and_copy(cursor,
-				       ufb,
-				       buffer,
-				       buf_byte_stride);
+		if (evdi_cursor_compose_and_copy(cursor,
+						 efb,
+						 buffer,
+						 buf_byte_stride))
+			EVDI_ERROR("Failed to blend cursor\n");
+
 		evdi_cursor_unlock(cursor);
 	}
-	return ret;
 }
 
 #define painter_lock(painter)                           \
@@ -213,8 +224,7 @@ u8 *evdi_painter_get_edid_copy(struct evdi_device *evdi)
 			memcpy(block,
 			       evdi->painter->edid,
 			       evdi->painter->edid_length);
-			EVDI_DEBUG("(dev=%d) %02x %02x %02x\n", evdi->dev_index,
-				   block[0], block[1], block[2]);
+			EVDI_DEBUG("(dev=%d) EDID valid\n", evdi->dev_index);
 		}
 	}
 	painter_unlock(evdi->painter);
@@ -305,7 +315,6 @@ void evdi_painter_send_cursor_set(struct evdi_painter *painter,
 		}
 		evdi_cursor_unlock(cursor);
 
-
 		event->base.event = &event->cursor_set.base;
 		event->base.file_priv = painter->drm_filp;
 #if KERNEL_VERSION(4, 8, 0) > LINUX_VERSION_CODE
@@ -422,6 +431,51 @@ static void evdi_painter_send_mode_changed(
 	}
 }
 
+int evdi_painter_get_num_dirts(struct evdi_device *evdi)
+{
+	struct evdi_painter *painter = evdi->painter;
+	int num_dirts;
+
+	if (painter == NULL) {
+		EVDI_WARN("Painter is not connected!");
+		return 0;
+	}
+
+	painter_lock(painter);
+
+	num_dirts = painter->num_dirts;
+
+	painter_unlock(painter);
+
+	return num_dirts;
+}
+
+struct drm_clip_rect evdi_painter_framebuffer_size(
+	struct evdi_painter *painter)
+{
+	struct drm_clip_rect rect = {0, 0, 0, 0};
+	struct evdi_framebuffer *efb = NULL;
+
+	if (painter == NULL) {
+		EVDI_WARN("Painter is not connected!");
+		return rect;
+	}
+
+	painter_lock(painter);
+	efb = painter->scanout_fb;
+	if (!efb) {
+		EVDI_DEBUG("Scanout buffer not set.");
+		goto unlock;
+	}
+	rect.x1 = 0;
+	rect.y1 = 0;
+	rect.x2 = efb->base.width;
+	rect.y2 = efb->base.height;
+unlock:
+	painter_unlock(painter);
+	return rect;
+}
+
 void evdi_painter_mark_dirty(struct evdi_device *evdi,
 			     const struct drm_clip_rect *dirty_rect)
 {
@@ -429,10 +483,16 @@ void evdi_painter_mark_dirty(struct evdi_device *evdi,
 	struct evdi_framebuffer *efb = NULL;
 	struct evdi_painter *painter = evdi->painter;
 
+	if (painter == NULL) {
+		EVDI_WARN("Painter is not connected!");
+		return;
+	}
+
 	painter_lock(evdi->painter);
 	efb = evdi->painter->scanout_fb;
 	if (!efb) {
-		EVDI_WARN("Skip clip rect. Scanout buffer not set.\n");
+		EVDI_DEBUG("(dev=%d) Skip clip rect. Scanout buffer not set.\n",
+			   evdi->dev_index);
 		goto unlock;
 	}
 
@@ -452,15 +512,26 @@ void evdi_painter_mark_dirty(struct evdi_device *evdi,
 	memcpy(&painter->dirty_rects[painter->num_dirts], &rect, sizeof(rect));
 	painter->num_dirts++;
 
-	if (painter->was_update_requested) {
-		evdi_painter_send_update_ready(painter);
-		painter->was_update_requested = false;
-	}
-
 unlock:
 	painter_unlock(evdi->painter);
 }
 
+void evdi_painter_send_update_ready_if_needed(struct evdi_device *evdi)
+{
+	struct evdi_painter *painter = evdi->painter;
+
+	if (painter) {
+		painter_lock(painter);
+		if (painter->was_update_requested) {
+			evdi_painter_send_update_ready(painter);
+			painter->was_update_requested = false;
+		}
+		painter_unlock(painter);
+	} else {
+		EVDI_WARN("Painter does not exist!");
+	}
+}
+
 void evdi_painter_dpms_notify(struct evdi_device *evdi, int mode)
 {
 	struct evdi_painter *painter = evdi->painter;
@@ -487,30 +558,55 @@ void evdi_painter_crtc_state_notify(struct evdi_device *evdi, int state)
 	}
 }
 
+static void evdi_log_pixel_format(uint32_t pixel_format)
+{
+#if KERNEL_VERSION(4, 10, 0) <= LINUX_VERSION_CODE
+	struct drm_format_name_buf format_name;
+
+	drm_get_format_name(pixel_format, &format_name);
+	EVDI_DEBUG("pixel format %s\n", format_name.str);
+#elif KERNEL_VERSION(4, 9, 0) <= LINUX_VERSION_CODE
+	char *format_name = drm_get_format_name(pixel_format);
+
+	EVDI_DEBUG("pixel format %s\n", format_name);
+	kfree(format_name);
+#else
+	const char *format_name = drm_get_format_name(pixel_format);
+
+	EVDI_DEBUG("pixel format %s\n", format_name);
+#endif
+}
+
 void evdi_painter_mode_changed_notify(struct evdi_device *evdi,
-				      struct drm_framebuffer *fb,
 				      struct drm_display_mode *new_mode)
 {
 	struct evdi_painter *painter = evdi->painter;
+	struct drm_framebuffer *fb = &painter->scanout_fb->base;
+	int bits_per_pixel;
+	uint32_t pixel_format;
+
+	if (fb == NULL)
+		return;
 
 #if KERNEL_VERSION(4, 11, 0) > LINUX_VERSION_CODE
-	int bits_per_pixel = fb->bits_per_pixel;
-	uint32_t pixel_format = fb->pixel_format;
+	bits_per_pixel = fb->bits_per_pixel;
+	pixel_format = fb->pixel_format;
 #else
-	int bits_per_pixel = fb->format->cpp[0] * 8;
-	uint32_t pixel_format = fb->format->format;
+	bits_per_pixel = fb->format->cpp[0] * 8;
+	pixel_format = fb->format->format;
 #endif
 
-	EVDI_DEBUG(
-		"(dev=%d) Notifying mode changed: %dx%d@%d; bpp %d; ",
-		evdi->dev_index, new_mode->hdisplay, new_mode->vdisplay,
-		drm_mode_vrefresh(new_mode), bits_per_pixel);
-	EVDI_DEBUG("pixel format %d\n", pixel_format);
+	EVDI_DEBUG("(dev=%d) Notifying mode changed: %dx%d@%d; bpp %d; ",
+		   evdi->dev_index, new_mode->hdisplay, new_mode->vdisplay,
+		   drm_mode_vrefresh(new_mode), bits_per_pixel);
+	evdi_log_pixel_format(pixel_format);
 
 	evdi_painter_send_mode_changed(painter,
 				       new_mode,
 				       bits_per_pixel,
 				       pixel_format);
+	if (painter)
+		painter->needs_full_modeset = false;
 }
 
 static int
@@ -523,7 +619,9 @@ evdi_painter_connect(struct evdi_device *evdi,
 	struct edid *new_edid = NULL;
 	int expected_edid_size = 0;
 
-	EVDI_CHECKPT();
+	EVDI_DEBUG("(dev=%d) Process is trying to connect\n",
+		   evdi->dev_index);
+	evdi_log_process();
 
 	if (edid_length < sizeof(struct edid)) {
 		EVDI_ERROR("Edid length too small\n");
@@ -567,6 +665,7 @@ evdi_painter_connect(struct evdi_device *evdi,
 	painter->edid_length = edid_length;
 	painter->edid = new_edid;
 	painter->is_connected = true;
+	painter->needs_full_modeset = true;
 
 	painter_unlock(painter);
 
@@ -575,8 +674,6 @@ evdi_painter_connect(struct evdi_device *evdi,
 
 	drm_helper_hpd_irq_event(evdi->ddev);
 
-	drm_helper_resume_force_mode(evdi->ddev);
-
 	return 0;
 }
 
@@ -590,19 +687,12 @@ static int evdi_painter_disconnect(struct evdi_device *evdi,
 	painter_lock(painter);
 
 	if (file != painter->drm_filp) {
-		EVDI_VERBOSE
-		    ("(dev=%d) An unknown connection to %p tries to close us",
-		     evdi->dev_index, file);
-		EVDI_VERBOSE(" - ignoring\n");
-
 		painter_unlock(painter);
 		return -EFAULT;
 	}
 
-	evdi_painter_set_new_scanout_buffer(evdi, NULL);
-
 	if (painter->scanout_fb) {
-		drm_framebuffer_unreference(&painter->scanout_fb->base);
+		drm_framebuffer_put(&painter->scanout_fb->base);
 		painter->scanout_fb = NULL;
 	}
 
@@ -654,6 +744,10 @@ int evdi_painter_connect_ioctl(struct drm_device *drm_dev, void *data,
 		else
 			ret = evdi_painter_disconnect(evdi, file);
 
+		if (ret) {
+			EVDI_WARN("(dev=%d)(pid=%d) disconnect failed\n",
+				  evdi->dev_index, (int)task_pid_nr(current));
+		}
 		return ret;
 	}
 	EVDI_WARN("Painter does not exist!");
@@ -666,93 +760,108 @@ int evdi_painter_grabpix_ioctl(struct drm_device *drm_dev, void *data,
 	struct evdi_device *evdi = drm_dev->dev_private;
 	struct evdi_painter *painter = evdi->painter;
 	struct drm_evdi_grabpix *cmd = data;
-	struct drm_framebuffer *fb = NULL;
 	struct evdi_framebuffer *efb = NULL;
-	int err = 0;
+	struct drm_clip_rect dirty_rects[MAX_DIRTS];
+	int err;
 
 	EVDI_CHECKPT();
 
+	if (cmd->mode != EVDI_GRABPIX_MODE_DIRTY) {
+		EVDI_ERROR("Unknown command mode\n");
+		return -EINVAL;
+	}
+
+	if (cmd->num_rects < 1) {
+		EVDI_ERROR("No space for clip rects\n");
+		return -EINVAL;
+	}
+
 	if (!painter)
 		return -ENODEV;
 
 	painter_lock(painter);
 
-	efb = painter->scanout_fb;
-
-	if (!efb) {
-		EVDI_ERROR("Scanout buffer not set\n");
-		err = -EAGAIN;
-		goto unlock;
-	}
-
 	if (painter->was_update_requested) {
 		EVDI_WARN("(dev=%d) Update ready not sent,",
 			  evdi->dev_index);
 		EVDI_WARN(" but pixels are grabbed.\n");
 	}
 
-	fb = &efb->base;
+	if (painter->num_dirts < 0) {
+		err = -EAGAIN;
+		goto err_painter;
+	}
+
+	merge_dirty_rects(&painter->dirty_rects[0],
+			  &painter->num_dirts);
+	if (painter->num_dirts > cmd->num_rects)
+		collapse_dirty_rects(&painter->dirty_rects[0],
+				     &painter->num_dirts);
+
+	cmd->num_rects = painter->num_dirts;
+	memcpy(dirty_rects, painter->dirty_rects,
+	       painter->num_dirts * sizeof(painter->dirty_rects[0]));
+
+	efb = painter->scanout_fb;
+
+	if (!efb) {
+		EVDI_ERROR("Scanout buffer not set\n");
+		err = -EAGAIN;
+		goto err_painter;
+	}
+
+	painter->num_dirts = 0;
+
+	drm_framebuffer_get(&efb->base);
+
+	painter_unlock(painter);
+
 	if (!efb->obj->vmapping) {
 		if (evdi_gem_vmap(efb->obj) == -ENOMEM) {
 			EVDI_ERROR("Failed to map scanout buffer\n");
 			err = -EFAULT;
-			goto unlock;
+			goto err_fb;
 		}
 		if (!efb->obj->vmapping) {
 			EVDI_ERROR("Inexistent vmapping\n");
 			err = -EFAULT;
-			goto unlock;
+			goto err_fb;
 		}
 	}
 
-	if (cmd->buf_width != fb->width ||
-		cmd->buf_height != fb->height) {
+	if (cmd->buf_width != efb->base.width ||
+		cmd->buf_height != efb->base.height) {
 		EVDI_ERROR("Invalid buffer dimension\n");
 		err = -EINVAL;
-		goto unlock;
+		goto err_fb;
 	}
 
-	if (cmd->num_rects < 1) {
-		EVDI_ERROR("No space for clip rects\n");
-		err = -EINVAL;
-		goto unlock;
+	if (copy_to_user(cmd->rects, dirty_rects,
+		cmd->num_rects * sizeof(cmd->rects[0]))) {
+		err = -EFAULT;
+		goto err_fb;
 	}
 
-	if (cmd->mode == EVDI_GRABPIX_MODE_DIRTY) {
-		if (painter->num_dirts < 0) {
-			err = -EAGAIN;
-			goto unlock;
-		}
-		merge_dirty_rects(&painter->dirty_rects[0],
-				  &painter->num_dirts);
-		if (painter->num_dirts > cmd->num_rects)
-			collapse_dirty_rects(&painter->dirty_rects[0],
-						 &painter->num_dirts);
+	err = copy_primary_pixels(efb,
+				  cmd->buffer,
+				  cmd->buf_byte_stride,
+				  cmd->num_rects,
+				  dirty_rects,
+				  cmd->buf_width,
+				  cmd->buf_height);
+	if (err == 0)
+		copy_cursor_pixels(efb,
+				   cmd->buffer,
+				   cmd->buf_byte_stride,
+				   evdi->cursor);
 
-		cmd->num_rects = painter->num_dirts;
+err_fb:
+	drm_framebuffer_put(&efb->base);
 
-		if (copy_to_user(cmd->rects, painter->dirty_rects,
-			cmd->num_rects * sizeof(cmd->rects[0])))
-			err = -EFAULT;
-		if (err == 0)
-			err = copy_primary_pixels(efb,
-					cmd->buffer,
-					cmd->buf_byte_stride,
-					painter->num_dirts,
-					painter->dirty_rects,
-					cmd->buf_width,
-					cmd->buf_height);
-		if (err == 0)
-			err = copy_cursor_pixels(efb,
-					cmd->buffer,
-					cmd->buf_byte_stride,
-					evdi->cursor);
+	return err;
 
-		painter->num_dirts = 0;
-	}
-unlock:
+err_painter:
 	painter_unlock(painter);
-
 	return err;
 }
 
@@ -792,9 +901,9 @@ int evdi_painter_init(struct evdi_device *dev)
 	dev->painter = kzalloc(sizeof(*dev->painter), GFP_KERNEL);
 	if (dev->painter) {
 		mutex_init(&dev->painter->lock);
-		mutex_init(&dev->painter->new_scanout_fb_lock);
 		dev->painter->edid = NULL;
 		dev->painter->edid_length = 0;
+		dev->painter->needs_full_modeset = true;
 		return 0;
 	}
 	return -ENOMEM;
@@ -816,52 +925,31 @@ void evdi_painter_cleanup(struct evdi_device *evdi)
 	}
 }
 
-/*
- * This can be called from multiple threads so we need to lock during
- * *new_scanout_fb* assignment.
- * It is called from *evdi_crtc_page_flip* which must return immediately.
- * If we lock here whole painter object it will interfere with grab_pics
- * ioctl (which can take some time).
- * Because of that we lock only on the *new_scanout_fb*.
- */
-void evdi_painter_set_new_scanout_buffer(struct evdi_device *evdi,
-					 struct evdi_framebuffer *newfb)
+void evdi_painter_set_scanout_buffer(struct evdi_device *evdi,
+				     struct evdi_framebuffer *newfb)
 {
 	struct evdi_painter *painter = evdi->painter;
 	struct evdi_framebuffer *oldfb = NULL;
 
 	if (newfb)
-		drm_framebuffer_reference(&newfb->base);
-
-	mutex_lock(&painter->new_scanout_fb_lock);
-	oldfb = painter->new_scanout_fb;
-	painter->new_scanout_fb = newfb;
-	mutex_unlock(&painter->new_scanout_fb_lock);
-
-	if (oldfb)
-		drm_framebuffer_unreference(&oldfb->base);
-}
-
-void evdi_painter_commit_scanout_buffer(struct evdi_device *evdi)
-{
-	struct evdi_painter *painter = evdi->painter;
-	struct evdi_framebuffer *newfb = NULL;
-	struct evdi_framebuffer *oldfb = NULL;
+		drm_framebuffer_get(&newfb->base);
 
 	painter_lock(painter);
-	mutex_lock(&painter->new_scanout_fb_lock);
-
-	newfb = painter->new_scanout_fb;
-
-	if (newfb)
-		drm_framebuffer_reference(&newfb->base);
 
 	oldfb = painter->scanout_fb;
 	painter->scanout_fb = newfb;
 
-	mutex_unlock(&painter->new_scanout_fb_lock);
 	painter_unlock(painter);
 
 	if (oldfb)
-		drm_framebuffer_unreference(&oldfb->base);
+		drm_framebuffer_put(&oldfb->base);
+}
+
+bool evdi_painter_needs_full_modeset(struct evdi_device *evdi)
+{
+	struct evdi_painter *painter = evdi->painter;
+
+	if (painter)
+		return painter->needs_full_modeset;
+	return false;
 }
